use pasetors::{
    claims::{Claims, ClaimsValidationRules},
    paserk::FormatAsPaserk,
    footer::Footer,
    keys::{Generate, AsymmetricKeyPair, AsymmetricSecretKey, AsymmetricPublicKey},
    public, Public, version4::V4,
    token::{UntrustedToken, TrustedToken},
};
use core::convert::TryFrom;
use std::collections::HashMap;

use serde::{Serialize,Deserialize,Deserializer,Serializer};
use std::io;
use std::io::prelude::*;
use std::fs::File;

type BxDynResult<T> = std::result::Result<T, Box<dyn std::error::Error>>;

fn read_byte_file(path: &str) -> io::Result<Vec<u8>> {
    let mut f = File::open(path)?;
    let mut buffer = Vec::new();
    // read the whole file
    f.read_to_end(&mut buffer)?;
    Ok(buffer)
}
fn read_file(path: &str) -> io::Result<String> {
    let content = std::fs::read_to_string(path)?;
    Ok(content)
}
fn default_config_pasetoken_string() -> String {
  String::from("")
}
fn default_config_pasetoken_footer() -> Footer {
  Footer::new()
}
fn serialize_config_pasetoken_footer<S>(_f: &Footer, serializer: S) -> Result<S::Ok, S::Error>
where S: Serializer {
    //let buf = String::deserialize(deserializer)?;
    let res = serializer.serialize_str("")?;
    Ok(res)
}
fn deserialize_config_pasetoken_footer<'de, D>(deserializer: D) -> Result<Footer, D::Error>
where D: Deserializer<'de> {
    let _buf = String::deserialize(deserializer)?;
    Ok(Footer::new())
}
/// Struct to collect settings for PaSeToken, `footer` will be loaded from `new` call 
/// From `serde` use `from_content` with content-file it will collect settings and load `footer`
/// To create a `PaSeToken` object use `pasetoken` 
#[derive(Clone, Serialize, Debug,Deserialize, Default)]
pub struct ConfigPaSeToken {
    pub public_path: String,
        #[serde(default = "default_config_pasetoken_string")]
    pub public_data: String,
    pub secret_path: String,
        #[serde(default = "default_config_pasetoken_string")]
    pub secret_data: String,
    pub is_bin: bool,
    pub assert_val: String,
    pub map_footer: HashMap::<String,String>,
        #[serde(default = "default_config_pasetoken_footer",
            deserialize_with = "deserialize_config_pasetoken_footer", 
            serialize_with = "serialize_config_pasetoken_footer"
        )]
    pub footer: Footer,
    pub data: HashMap::<String,String>,
    pub expire: bool,
}
#[allow(unused)]
impl ConfigPaSeToken {
    pub fn new(
        public_path: String,
        secret_path: String,
        is_bin: bool,
        assert_val: String,
        map_footer: HashMap::<String,String>,
        data: HashMap::<String,String>,
        expire: bool,
    ) -> Self {
        Self {
            public_path: public_path.to_owned(),
            public_data: String::from(""),
            secret_path: secret_path.to_owned(),
            secret_data: String::from(""),
            is_bin,
            assert_val: assert_val.to_owned(),
            map_footer: map_footer.to_owned(),
            data: data.to_owned(),
            footer: Self::make_footer(map_footer.to_owned()).unwrap_or(Footer::new()),
            expire, 
        }
    }
    pub fn load_data(&self) -> (String,String){
        (
            read_file(&self.public_path).unwrap_or(String::from("")),
            read_file(&self.secret_path).unwrap_or(String::from("")),
        )
    }
    pub fn pasetoken(&self) ->  BxDynResult<PaSeToken> {
        Ok(PaSeToken::new(self, &self.assert_val,&self.footer)?)
    }
    pub fn make_footer(map_footer: HashMap<String,String>) -> BxDynResult<Footer> {
       if map_footer.len() > 0 {
            let mut data_footer = Footer::new();
            for (key, value) in map_footer.iter() {
                data_footer.add_additional(key, value.as_str())?;
            } 
            Ok(data_footer)
        }  else {
            Ok(Footer::new())
        }
    }
    pub fn from_content(content: &str, frmt: &str) ->  BxDynResult<Self> {
        let mut cfg: ConfigPaSeToken = match frmt {
           "toml" => toml::from_str(&content).unwrap_or_else(|e| {
                        eprintln!("Error parse config parsetoken: {}",e);
                        ConfigPaSeToken::default()
                 }),
            _ => serde_json::from_str(&content).unwrap_or_else(|e| {
                    eprintln!("Error parse config parsetoken: {}",e);
                    ConfigPaSeToken::default()
                }),
         };
        cfg.footer = Self::make_footer(cfg.map_footer.to_owned()).unwrap_or(Footer::new());
        (cfg.public_data, cfg.secret_data) = cfg.load_data();
        Ok(cfg)
    }
    pub fn generate_token(&self, path: &str, data: &HashMap::<String,String>, expire: bool) -> BxDynResult<String> {
        //dbg!(data);
        let pasetoken = self.pasetoken()?;
        let token = pasetoken.generate(data, expire)?;
        if ! path.is_empty() {
           std::fs::write(path, &token)?; 
        }
        Ok(token)
    }
    pub fn token_from_file_defs(def_path: &str, out_path: &str) -> BxDynResult<String> {
        let content = std::fs::read_to_string(&def_path)?;
        let frmt = if def_path.ends_with(".toml") { "toml" } else { "json" };
        let cfg_pasetoken = ConfigPaSeToken::from_content(&content, frmt)?;
        cfg_pasetoken.generate_token(out_path, &cfg_pasetoken.data, cfg_pasetoken.expire) 
    }
}
pub struct PaSeToken<'a> {
    claims: Claims,
    kp: AsymmetricKeyPair<V4>,
    footer: Option<&'a Footer>,
    assert: Option<&'a[u8]>,
}
/// PaSeToken manager: new from `ConfigPaSeToken` + `Footer` .  
#[allow(unused)]
impl<'a> PaSeToken<'a> {
    pub fn new(config: &'a ConfigPaSeToken, assert_val: &'a str, op_footer: &'a Footer) -> BxDynResult<Self> {
        // Setup the default claims, which include `iat` and `nbf` as the current time and `exp` of one hour.
        // Generate claims 
        // Add a custom `data` claim as well.
        let mut claims = Claims::new()?;
        for (key, value) in config.data.iter() {
            claims.add_additional(key, value.to_owned())?;
        }
        if ! config.expire {
            claims.non_expiring();
        }
        // Generate the keys if are empty
        let kp = if config.public_path.is_empty() || config.secret_path.is_empty() {
          AsymmetricKeyPair::<V4>::generate()?
        } else {
           if config.is_bin {
                let (loaded_public, loaded_secret) = PaSeToken::from_files_bin(&config.public_path, &config.secret_path)?;
                let public: &[u8] = &loaded_public;
                let secret: &[u8] = &loaded_secret;
                AsymmetricKeyPair::<V4> { 
                    public: AsymmetricPublicKey::<V4>::from(public)?,
                    secret: AsymmetricSecretKey::<V4>::from(secret)?,
                }
           } else {
                let loaded_public: String;
                let loaded_secret: String;
                if config.public_data.is_empty() || config.secret_data.is_empty() {
                    (loaded_public,loaded_secret) = PaSeToken::from_files(&config.public_path, &config.secret_path)?;
                } else {
                    loaded_public = config.public_data.to_owned();
                    loaded_secret = config.secret_data.to_owned();
                };
                AsymmetricKeyPair::<V4> { 
                    public: AsymmetricPublicKey::<V4>::try_from(loaded_public.as_str())?,
                    secret: AsymmetricSecretKey::<V4>::try_from(loaded_secret.as_str())?,
                }
           }
        };
        let assert: Option<&'a[u8]> = if assert_val.len() > 0 {
              Some(config.assert_val.as_bytes())
        } else { 
            None
        };
        let footer = if config.map_footer.len() > 0 {
            Some(op_footer)
        } else {
            None
        };
        Ok(Self {
            claims,
            kp,
            footer: footer.to_owned(),
            assert: assert.to_owned(),
        }) 
    }
    /// Get public key as String 
    pub fn parserk_public(&self) -> String {
        let mut paserk = String::new();
        match self.kp.public.fmt(&mut paserk) {
            Ok(_) => paserk,
            Err(_) => String::from(";"),
        }
    }
    /// Get secret key as String 
    pub fn parserk_secret(&self) -> String {
        let mut paserk = String::new();
        match self.kp.secret.fmt(&mut paserk) {
            Ok(_) => paserk,
            Err(_) => String::from(""),
        }
    }
    /// Load keys from binary files
    pub fn from_files_bin(public_path: &str, secret_path: &str) -> BxDynResult<(Vec<u8>,Vec<u8>)> {
        let public: Vec<u8> = read_byte_file(&public_path)?;
        let secret: Vec<u8> = read_byte_file(&secret_path)?;
        Ok((public, secret))
    }
    /// Write keys to binary files
    pub fn to_path_bin(&self, public_path: &str, secret_path: &str ) -> BxDynResult<()> {
        if ! public_path.is_empty() {
            let public = self.kp.public.as_bytes();
            std::fs::write(public_path, public)?;
        }
        if ! secret_path.is_empty() {
            let secret = self.kp.secret.as_bytes();
            std::fs::write(secret_path, secret)?;
        }
        Ok(())
    }
    /// Load keys from text files
    pub fn from_files(public_path: &str, secret_path: &str) -> BxDynResult<(String,String)> {
        let public: String = read_file(&public_path)?;
        let secret: String = read_file(&secret_path)?;
        Ok((public, secret))
    }
    /// Write keys to text files
    pub fn to_path(&self, public_path: &str, secret_path: &str ) -> BxDynResult<()> {
        if ! public_path.is_empty() {
            let public = self.parserk_public();
            std::fs::write(public_path, public)?;
        }
        if ! secret_path.is_empty() {
            let secret = self.parserk_secret();
            std::fs::write(secret_path, secret)?;
        }
        Ok(())
    }
    // Generate a Token adding some data form `claims` with expire (true (1 hour)/false (forever))
    pub fn generate(&self, data: &HashMap::<String,String>, expire: bool) -> BxDynResult<String> {
        let claims = if data.is_empty() {
            self.claims.to_owned()
        } else {
            let mut claims = Claims::new()?;
            for (key, value) in data.iter() {
              claims.add_additional(key, value.to_owned())?;
            }
            if ! expire {
                claims.non_expiring();
            }
            claims
        };
        let pase_token = public::sign(
            &self.kp.secret,
            &claims,
            self.footer,
            self.assert
        )?;
        // println!("Pub Token: {}", &pase_token);
        Ok(pase_token)
    }
    /// Get a `TrustedToken` from string and apply validation rules.
    pub fn trusted(&self, token: &str, expire: bool) -> BxDynResult<TrustedToken> {
        // Decide how we want to validate the claims after verifying the token itself.
        // The default verifies the `nbf`, `iat` and `exp` claims. `nbf` and `iat` are always
        // expected to be present.
        // NOTE: Custom claims, defined through `add_additional()`, are not validated. This must be done
        // manually.
        let mut validation_rules = ClaimsValidationRules::new();
        if ! expire {
            validation_rules.allow_non_expiring();
        }
        let untrusted_token = UntrustedToken::<Public, V4>::try_from(token)?;
        let trusted_token = public::verify(
            &self.kp.public,
            &untrusted_token,
            &validation_rules,
            self.footer,
            self.assert
        )?;
        Ok(trusted_token)
    } 
    /// Get `claims`  with some additional data
    fn get_claims(&self, data: &HashMap<String,String>, expire: bool) -> BxDynResult<Claims> {
        if data.len() == 0 {
            Ok(self.claims.to_owned())
        } else {
            let mut claims = Claims::new()?;
            for (key, value) in data.iter() {
                claims.add_additional(key, value.to_owned())?;
            }
            if ! expire {
                claims.non_expiring();
            }
           Ok(claims)
        }
    }
    /// Validate example over `TrustedToken` and data in `claims`
    /// implement this by looking at `claims` from `trusted_token.payload_claims`
    pub fn validate(&self, trusted_token: TrustedToken, data: &HashMap::<String,String>, expire: bool) -> bool {
        let source_claims = if data.len() == 0 {
            self.claims.to_owned()
        } else {
            self.get_claims(data, expire).unwrap_or(self.claims.to_owned())
        };
        if let Some(claims) = trusted_token.payload_claims() {
            println!("{:?}", claims.get_claim("data"));
            println!("{:?}", claims.get_claim("iat"));
            if let Some(val) = claims.get_claim("data") {
                if let Some(src_data) = source_claims.get_claim("data") {
                    val == src_data
                } else {
                    false
                }
            } else {
                false
            }
        } else {
            false
        }
    }
}
