/// Tests using `test-log` dependency to capture `log` macros.
/// use command:  cargo test -- --nocapture 
#[cfg(test)]
mod tests {
    use log::{info,error};
    use test_log::test;
    use crate::pasetoken::ConfigPaSeToken;
    use std::collections::HashMap;
    type BxDynResult<T> = std::result::Result<T, Box<dyn std::error::Error>>;
    pub fn test_token<'a>(
            public_path: &str,
            secret_path: &str,
            mode: bool,
            data: &HashMap<String,String>,
            file_public: &str,
            file_secret: &str
        ) -> BxDynResult<bool>  {
        let expire = true;
        let config_pasetoken: ConfigPaSeToken = ConfigPaSeToken::new(
            public_path.to_owned(),
            secret_path.to_owned(),
            mode,
            String::from("implicit assertion"),
            HashMap::new(),
            data.to_owned(),
            false
        );
        let pase_token = config_pasetoken.pasetoken()?;
        let token = pase_token.generate(&data, expire).unwrap_or_else(|e|{
            error!("Error generate token: {}",e);
            String::from("")
        });
        println!("token: {}", &token);
        let is_valid = match pase_token.trusted(&token, expire) {
            Ok(trusted_token) => pase_token.validate(trusted_token, &data, expire),
            Err(_e) => false,
        };
        info!("Is valid: {}", &is_valid);
        // let h: PaSeToken = pase_token.try_into().unwrap();
        if mode {
            pase_token.to_path_bin(file_public, file_secret)?;
        } else {
            info!("parserk public: {}", pase_token.parserk_public());
            info!("parserk secret: {}", pase_token.parserk_secret());
            pase_token.to_path(file_public, file_secret)?;
        }
        Ok(is_valid)
    }
    pub fn clean_files(list: Vec<String>) {
        for file in list {
            match std::fs::remove_file(&file) {
                Ok(_) => info!("File {} Removed", &file),
                Err(e) => error!("Error removing file {}: {}", &file,e),
            };
        }
    }
    #[test]
    pub fn chek_token_bin<'a>() {
        let mut data = HashMap::new();
        let public_path = String::from("/tmp/bin_public_bin");
        let secret_path = String::from("/tmp/bin_secret_bin");
        data.insert("data".to_string(),"A public, signed message".to_string());
        let mode = true; // to save files as binary
        match test_token("","", mode,  &data, &public_path, &secret_path) {
            Ok(_) => 
                match test_token(&public_path, &secret_path, mode, &data, "", "") {
                    Ok(_) => assert!(true),
                    Err(e) => {
                        error!("Error: {}",e);
                        assert!(false);
                    },
                },
            Err(e) => {
                error!("Error: {}",e);
                assert!(false);
            },
        }
        clean_files(vec![public_path, secret_path]);
    }
    #[test]
    pub fn chek_token_val() {
        let mut data = HashMap::new();
        let public_path = String::from("/tmp/public");
        let secret_path = String::from("/tmp/secret");
        data.insert("data".to_string(),"A public, signed message".to_string());
        let mode = false;  // to save files as parserk 
        match test_token("","", mode,  &data, &public_path, &secret_path) {
            Ok(_) => 
                match test_token(&public_path, &secret_path, mode, &data, "", "") {
                    Ok(_) => assert!(true),
                    Err(e) => {
                        error!("Error: {}",e);
                        assert!(false);
                    },
                },
            Err(e) => {
                error!("Error: {}",e);
                assert!(false);
            },
        }
        clean_files(vec![public_path, secret_path]);
    }
    #[test]
    pub fn serde_json() {
        let mut data = HashMap::new();
        let public_path = String::from("/tmp/public");
        let secret_path = String::from("/tmp/secret");
        data.insert("data".to_string(),"A public, signed message".to_string());
        let mode = false;  // to save files as parserk 
        let config_pasetoken: ConfigPaSeToken = ConfigPaSeToken::new(
            public_path.to_owned(),
            secret_path.to_owned(),
            mode,
            String::from("implicit assertion"),
            HashMap::new(),
            data.to_owned(),
            false
        );
        match serde_json::to_string(&config_pasetoken) {
            Ok(res) => {
                println!("{}",&res);
                match serde_json::from_str::<ConfigPaSeToken>(&res) {
                    Ok(_) => assert!(true),
                    Err(e) => {
                            error!("Error: {}",e);
                            assert!(false);
                        },
                    };
            },
            Err(e) => {
                error!("Error: {}",e);
                assert!(false);
            },
        }
    }
    #[test]
    pub fn from_content() {
        let mut data = HashMap::new();
        let public_path = String::from("/tmp/public");
        let secret_path = String::from("/tmp/secret");
        data.insert("data".to_string(),"A public, signed message".to_string());
        let mode = false;  // to save files as parserk 
        let config_pasetoken: ConfigPaSeToken = ConfigPaSeToken::new(
            public_path.to_owned(),
            secret_path.to_owned(),
            mode,
            String::from("implicit assertion"),
            HashMap::new(),
            data.to_owned(),
            false
        );
        match serde_json::to_string(&config_pasetoken) {
            Ok(res) => 
                match ConfigPaSeToken::from_content(&res, "json") {
                    Ok(_) => assert!(true),
                    Err(e) => {
                            error!("Error: {}",e);
                            assert!(false);
                        },
                    },
            Err(e) => {
                error!("Error: {}",e);
                assert!(false);
            },
        }
    }
    #[test]
    pub fn token_from_file_defs() {
        let config_pasetoken: ConfigPaSeToken = ConfigPaSeToken::new(
            String::from(""),
            String::from(""),
            false,
            String::from("implicit assertion"),
            HashMap::new(),
            HashMap::new(),
            false
        );
        match serde_json::to_string(&config_pasetoken) {
            Ok(content) => 
                match ConfigPaSeToken::from_content(&content, "json") {
                    Ok(cfg) => {
                        let token = cfg.generate_token(
                            "",
                            &cfg.data,
                            cfg.expire
                        ).unwrap_or_else(|e|{
                            eprintln!("Token generation error: {}",e);
                            String::from("")
                        });
                        assert_eq!(token.is_empty(),false);
                    },
                    Err(e) => {
                        error!("Error config from content: {}",e);
                        assert!(false);
                    },
                },
            Err(e) => {
                error!("Error paseteoken config: {}",e);
                assert!(false);
            },
        };   
        
    }
}